import pandas as pd

def update_blocks_and_reset_prev(blocks, prev, row, 
        features=['clone_id', 'chr', 'start', 'end', 'state']):
    if prev['chr']: # not null
        blocks = pd.concat([blocks, prev.to_frame().T])
    prev = pd.Series({f: row[f] for f in features}, index=features)
    return blocks, prev
    
def merge_cn_segments(cn, merge_gap=100000):
    #in_pp = '/juno/work/shah/isabl_data_lake/analyses/37/74/23774/SPECTRUM-OV-081_S1_LEFT_OVARY_cn.csv'
    features = ['clone_id', 'chr', 'start', 'end', 'state']
    cn = cn[features].copy() # reduce features for the sake of memory

    # merge segments with same metric
    prev = pd.Series({f:None for f in features}, index=features) # init prev cn
    blocks = pd.DataFrame(columns=features) # init empty blocks

    for rix, row in cn.iterrows():
        if prev['chr'] != row['chr']:
            blocks, prev = update_blocks_and_reset_prev(blocks, prev, row)
            continue
        if (prev['state'] != row['state']):
            blocks, prev = update_blocks_and_reset_prev(blocks, prev, row)
            continue
        if (prev['state'] == row['state']) and abs(prev['end'] - row['start']) > merge_gap:
            blocks, prev = update_blocks_and_reset_prev(blocks, prev, row)
            continue
        # update prev block
        prev['end'] = row['end']

    blocks = pd.concat([blocks, prev.to_frame().T]) # add prev
    blocks.loc[:, ['start', 'end', 'state']] = blocks.loc[:, ['start', 'end', 'state']].astype(int)
    blocks['start'] = blocks['start'] # so segments don't overlap; w/o this !mixFlag[j] error occurs in MutationTimeR step
    return blocks.reset_index(drop=True)

def filter_sv_by_clone(sv, vaf_col_str='vaf_'):
    vaf_cols = [c for c in sv.columns if c.startswith(vaf_col_str)]
    clone_ids = [c.replace(vaf_col_str, '') for c in vaf_cols]
    clonal_sv = {}
    for clone_id in clone_ids:
        clonal = f'vaf_{clone_id}'
        others = [f'vaf_{c}' for c in clone_ids if c != clone_id]
        clonal_sv[clone_id] = sv[
            (sv[clonal] > 0) &
            (sv[others].sum(axis=1) == 0)
        ].reset_index(drop=True)
    return clonal_sv


from collections import Counter

import tqdm
import pandas as pd
import pysam

from .utils import get_secondaries, make_split_read_table, enumerate_breakpoints
from .datatypes import SplitAlignment, Breakpoint, BreakpointPair

def extract_split_alignments(reads, max_depth=500):
    alignments = []
    for i, read in enumerate(reads):
        if i >= max_depth:
            break
        secondary_list = get_secondaries(read)
        if len(secondary_list) < 1: # only consider reads with secondaries (chimeric reads)
            continue # return None
        strand = '-' if read.is_reverse else '+'
        tpos = read.pos + 1
        cigarstring = read.cigarstring
        qname = read.qname
        refname = read.reference_name
        alignment = SplitAlignment(cigarstring, qname, refname, tpos, strand)
        alignments.append(alignment)

        for _, secondary in enumerate(secondary_list):
            refname_s, tpos_s, strand_s, cigarstring_s, _, _ = secondary.split(',')
            tpos_s = int(tpos_s)
            alignment = SplitAlignment(cigarstring_s, qname, refname_s, tpos_s, strand_s)
            alignments.append(alignment)
    return alignments

def pull_breakpoints_from_reads_in_sv_regions(bam, tra, get_read_table=False, min_n_breakpoint=3, margin=10):
    complexes = []
    saved_qnames = set()
    canonical_chroms = [str(c) for c in range(1, 22+1)] + ['X', 'Y']
    canonical_chroms += ['chr'+c for c in canonical_chroms]
    read_info = pd.DataFrame()
    for tix, row in tqdm.tqdm(tra.iterrows(), total=tra.shape[0]):
        # row = tra.iloc[6] # breakpoint pair: SV
        chrom1 = row['chromosome_1']
        pos1 = row['position_1']
        chrom2 = row['chromosome_2']
        pos2 = row['position_2']
        reads1 = bam.fetch(chrom1, pos1-margin-1, pos1+margin)
        reads2 = bam.fetch(chrom2, pos2-margin-1, pos2+margin)
        for reads in [reads1, reads2]:
            alignments = extract_split_alignments(reads)
            df = make_split_read_table(alignments)
            df = df[~df['qname'].isin(saved_qnames)] # don't double-count breakpoints already saved
            read_info = pd.concat([read_info, df])
            df = df[df['chrom'].isin(canonical_chroms)] # remove segments in decoys
            _qnames = set(df['qname'].unique().tolist())
            saved_qnames.update(_qnames)
            bundle = make_brks_bundle(df)
            for brks in bundle:
                if len(brks) >= min_n_breakpoint:
                    complexes.append(brks)
    if get_read_table:
        return complexes, read_info
    return complexes

def make_brks_bundle(df):    
    bundle = []
    for qname, qdf in df.groupby('qname'):
        brks = enumerate_breakpoints(qdf)
        brks.qname = qname
        brks.info = {'sv':brks.tras, 'qname':brks.qname}
        brks.get_transitions(sort_transition=False)
        bundle.append(brks)
    return bundle

def get_breakpoint_support_from_bundle(complexes):
    breakpoint_support = Counter()
    for brks in complexes:
        for brk in brks:
            chrom, pos, ori = brk.chrom, brk.pos, brk.ori
            breakpoint_support[str(brk)] += 1
    return breakpoint_support

def map_similar_coordinate_to_higher_rank(complexes, breakpoint_support, margin=10):
    coord_map = {}
    coord_map_log = {}
    for cix, brks in enumerate(complexes):
        for bix, brk in enumerate(brks):
            chrom, pos, ori = brk.chrom, brk.pos, brk.ori 
            src_coord = str(brk)
            src_count = breakpoint_support[src_coord]
            max_count = src_count
            max_coord = src_coord
            for _pos in range(pos-margin, pos+margin):
                _coord = f'{chrom}:{_pos}:{ori}'
                if _coord in breakpoint_support:
                    _count = breakpoint_support[_coord]
                    if _count > max_count:
                        max_count = _count
                        max_coord = _coord
            coord_map[src_coord] = max_coord
            coord_map_log[src_coord] = (max_coord, src_count, max_count)
    return coord_map, coord_map_log

def fix_lower_support_coordinates(complexes, coord_map):
    for cix, brks in enumerate(complexes):
        for bix, brk in enumerate(brks):
            src_coord = str(brk)
            dst_coord = coord_map[src_coord]
            dst_chrom, dst_pos, dst_ori = dst_coord.split(':')
            dst_pos = int(dst_pos)
            complexes[cix][bix].pos = dst_pos
    return complexes

def normalize_sv_table(df, chrom1_col='chromosome_1', chrom2_col='chromosome_2', 
                      pos1_col='position_1', pos2_col='position_2', 
                      ori1_col='strand_1', ori2_col='strand_2'):
    df = df.copy()
    chroms = [str(c) for c in range(1, 22+1)]+['X', 'Y']
    chroms += ['chr'+c for c in chroms] # chr prefices
    chrom_map = dict(zip(chroms, range(len(chroms))))
    assert (~df[chrom1_col].isin(chroms)).sum() == 0, df[chrom1_col].unique()
    assert (~df[chrom2_col].isin(chroms)).sum() == 0, df[chrom2_col].unique()
    diff_chr = df[chrom1_col] != df[chrom2_col]
    flag_inverse = (df[chrom1_col].map(chrom_map) > df[chrom2_col].map(chrom_map))
    flag_inverse |= (df[chrom1_col]==df[chrom2_col]) & (df[pos1_col] > df[pos2_col])
    df.loc[flag_inverse, [chrom1_col, chrom2_col]] = df.loc[flag_inverse, [chrom2_col, chrom1_col]].values
    df.loc[flag_inverse, [pos1_col, pos2_col]] = df.loc[flag_inverse, [pos2_col, pos1_col]].values
    df.loc[flag_inverse, [ori1_col, ori2_col]] = df.loc[flag_inverse, [ori2_col, ori1_col]].values
    return df

def get_svtype(tra:BreakpointPair):
    translocation = 'TRA'
    inversion = 'INV'
    duplication = 'DUP'
    insertion = 'INS'
    deletion = 'DEL'
    chrom1 = tra.brk1.chrom
    chrom2 = tra.brk2.chrom
    if chrom1 != chrom2:
        return translocation
    else: # same chrom
        pos1, pos2 = tra.brk1.pos, tra.brk2.pos
        ori1, ori2 = tra.brk1.ori, tra.brk2.ori
        if pos1 > pos2:
            pos1, pos2 = pos2, pos1
            ori1, ori2 = ori2, ori1
        if ori1 == ori2:
            return inversion
        else:
            if ori1 == '+':
                if ori2 == '-':
                    return deletion
            elif ori1 == '-':
                if ori2 == '+':
                    return duplication
    raise ValueError(f'tra:{tra}')

def pull_sv_supporting_reads_from_bundle(sv, bundle):
    output_bundle = []
    sv_prop = {
        (sv['chromosome_1'], sv['position_1'], sv['strand_1']), 
        (sv['chromosome_2'], sv['position_2'], sv['strand_2']), 
    }
    for brks in bundle:
        brks.contains_sv_support = False
        for tra in brks.tras:
            tra.supports_sv = False
            brk1 = tra.brk1
            brk2 = tra.brk2
            tra_prop = {
                (brk1.chrom, brk1.pos, brk1.ori),
                (brk2.chrom, brk2.pos, brk2.ori),
            }
            if sv_prop == tra_prop:
                tra.supports_sv = True
                brks.contains_sv_support = True
        if brks.contains_sv_support:
            output_bundle.append(brks)
    return output_bundle

def find_presence_of_matching_sv(sv1, sv2, margin=50):
    match = []
    for i, (rix, row) in enumerate(sv2.iterrows()):
        _match = (
            (sv1['chromosome_1'] == row['chromosome_1']) &
            (sv1['chromosome_2'] == row['chromosome_2']) &
            (abs(sv1['position_1']-row['position_1']) < margin) &
            (abs(sv1['position_2']-row['position_2']) < margin) &
            (sv1['strand_1'] == row['strand_1']) &
            (sv1['strand_2'] == row['strand_2'])
        )
        if i == 0:
            match = _match
        else:
            match |= _match
    return match

def pull_breakpoints_from_bam_files(bam_paths, sv, get_read_table=False):
    complexes = []
    read_tables = pd.DataFrame()
    for bam_path in bam_paths:
        bam = pysam.AlignmentFile(bam_path)
        if get_read_table:
            _complex, read_table = pull_breakpoints_from_reads_in_sv_regions(bam, sv, get_read_table=get_read_table)
            read_tables = pd.concat([read_tables, read_table])
        else:
            _complex = pull_breakpoints_from_reads_in_sv_regions(bam, sv, get_read_table=get_read_table)
        complexes += _complex
    if get_read_table:
        return complexes, read_tables
    return complexes

def make_tumor_sv_table(complexes, sv, margin=10, get_support=True):
    breakpoint_support = get_breakpoint_support_from_bundle(complexes)
    coord_map, coord_map_log = map_similar_coordinate_to_higher_rank(complexes, breakpoint_support, margin=margin)
    complexes = fix_lower_support_coordinates(complexes, coord_map)
    # breakpoint_support_post = get_breakpoint_support_from_bundle(complexes)
    sv_cols = ['chromosome_1', 'position_1', 'strand_1', 'chromosome_2', 'position_2', 'strand_2', 'type']
    tumor_sv = pd.DataFrame(columns=sv_cols)
    for brks in complexes:
        for tra in brks.tras:
            svtype = get_svtype(tra)
            chrom1, chrom2 = tra.brk1.chrom, tra.brk2.chrom
            pos1, pos2 = tra.brk1.pos, tra.brk2.pos
            ori1, ori2 = tra.brk1.ori, tra.brk2.ori
            field = [chrom1, pos1, ori1, chrom2, pos2, ori2, svtype]
            tumor_sv.loc[tumor_sv.shape[0]] = field
    tumor_sv = normalize_sv_table(tumor_sv)
    ix_cols = ['chromosome_1', 'position_1', 'strand_1', 'chromosome_2', 'position_2', 'strand_2', 'type']
    if get_support:
        tumor_sv = tumor_sv.groupby(ix_cols).value_counts().reset_index()
        tumor_sv.rename(columns={'count':'support'}, inplace=True)
    tumor_sv['in_source'] = find_presence_of_matching_sv(tumor_sv, sv, margin=50)
    return tumor_sv

def get_normalized_sv(tra):
    chrom1, pos1, ori1 = tra.brk1.chrom, tra.brk1.pos, tra.brk1.ori
    chrom2, pos2, ori2 = tra.brk2.chrom, tra.brk2.pos, tra.brk2.ori
    if Breakpoint.chroms.index(chrom1) < Breakpoint.chroms.index(chrom1):
        chrom1, pos1, ori1, chrom2, pos2, ori2 = chrom2, pos2, ori2, chrom1, pos1, ori1
    elif chrom1 == chrom2:
        if pos1 > pos2:
            chrom1, pos1, ori1, chrom2, pos2, ori2 = chrom2, pos2, ori2, chrom1, pos1, ori1
    return [chrom1, pos1, ori1, chrom2, pos2, ori2]




import re
import pandas as pd
import numpy as np


class SplitAlignment:
    """
    Represents a split alignment of a sequencing read, encapsulating details about the alignment's location,
    orientation, and structure as described by its CIGAR string.

    This class is designed to parse and store information about a single alignment segment of a read, which
    can be part of a primary alignment or a secondary (chimeric) alignment. It includes methods to interpret
    the CIGAR string for further analysis of the alignment's properties, such as calculating the aligned
    length and identifying specific alignment features.
    """
    def __init__(self, cigarstring, read_name, refname, read_pos, strand):
        """
        Attributes:
        - cigarstring (str): The CIGAR string representing the alignment's match, insertion, deletion, 
            skipping, soft clipping, hard clipping, padding, sequence match, and sequence mismatch operations.
        - read_name (str): The name (query name) of the read.
        - refname (str): The name of the reference chromosome to which the read is aligned.
        - read_pos (int): The starting position of the alignment on the reference chromosome (0-based).
        - strand (str): The strand of the alignment, represented as '+' for the forward strand and '-' for
            the reverse strand.
        """
        self.read_name = read_name
        self.refname = refname
        self.cigarstring = cigarstring
        self.cigar_tuples = SplitAlignment.get_cigar_tuples(self.cigarstring)
        self.primary = None
        self.start = read_pos
        self.strand = strand
        self.match = 0

        self.aln_cols = ['qname', 'chrom', 'start', 'end', 'strand', 'clip1', 'match', 'clip2', 'pclip1']
        self.extract_cigar_field()
        
    def extract_cigar_field(self):
        split_ctags = {4, 5}
        query_consumers = {0, 1, 4, 7, 8}
        reference_consumers = {0, 2, 3, 7, 8}
        self.clip1, self.match, self.clip2 = 0, 0, 0
        flag_after_match = False
        chrom, pos, strand, cigar_tuples = self.refname, self.start, self.strand, self.cigar_tuples
        tpos = int(pos) # init reference pos
        for ctag, clen in cigar_tuples:
            if ctag in split_ctags:
                if not flag_after_match:
                    self.clip1 += clen
                else:
                    self.clip2 += clen
            elif ctag in query_consumers:
                flag_after_match = True
                self.match += clen
            if ctag in reference_consumers:
                flag_after_match = True
                tpos += clen
                continue
        self.end = tpos
        pclip1 = self.clip1 if strand == '+' else self.clip2
        self.pclip1 = pclip1 # positively corrected first clip length

    @staticmethod
    def get_cigar_tuples(cigarstring):
        """
        Returns a cigar tuple from a CIGAR string
        """
        cigar_types = 'MIDNSHP=X'
        converter = {
            cigar_types[i]:i for i in range(len(cigar_types))
        }
        cigar_tuples = []
        for match in re.finditer('(\d+)([A-Z\=])', cigarstring):
            clen, ctype = match.groups()
            cigar_tuple = (converter[ctype], int(clen))
            cigar_tuples.append(cigar_tuple)
        return cigar_tuples


class BreakpointChain(list):
    def __init__(self, brks_iterable):
        super().__init__(brks_iterable)
        self.tras = []
        self.segs = []

    # enumeration of tras
    def get_transitions(self, sort_transition=False):
        if len(self) >= 2:
            ix_range = range(0, len(self), 2)
            for i in ix_range:
                brk1 = self[i]
                brk2 = self[i+1]
                if sort_transition:
                    if brk1 > brk2:
                        brk1, brk2 = brk2, brk1
                tra = BreakpointPair(brk1, brk2)
                self.tras.append(tra)

    # enumeration of segs
    def get_segments(self):
        ix_range = range(1, len(self)-1, 2)
        for i in ix_range:
            brk1 = self[i]
            brk2 = self[i+1]
            seg = BreakpointPair(brk1, brk2)
            self.segs.append(seg)


class Sample:
    long_samples = [ # class attribute
        "14472B_201",
        "14472B_202",
        "14472B_500",
        "14472B_501",
        "14472C_203",
        "14472C_204",
        "14472C_502",
        "14472D_101",
        "14472D_102",
        "14472D_103",
        "14472D_104",
        "14472D_205",
        "14472_100",
        "14472_300",
    ]
    sample2group = {s:s.split('_')[1][0] for s in long_samples}
    sample2short = {s:s.split('_')[1] for s in long_samples}
    
    def __init__(self, long_sample):
        self.long_sample = long_sample
        self.group = self.sample2group[self.long_sample]
        self.long_id = self.long_sample
        self.id = self.sample2short[self.long_sample]

    def __repr__(self):
        return f'{self.id}'


class Transitions:
    def __init__(self, df):
        self.df = df.copy()
        self.list = []
        self.get_list()

    def get_list(self):
        for qname, qdf in self.df.groupby('qname'):
            qdf.reset_index(drop=True, inplace=True)
            n_fragments = qdf.shape[0]
            if n_fragments <= 1: continue # don't count TRA when none
            prevdf = qdf.shift(1)
            for rix, row in qdf.iterrows():
                if rix == 0: continue
                prow = prevdf.loc[rix]
                chrom1, start1, end1, strand1 = prow['chrom'], int(prow['start']), int(prow['end']), prow['strand']
                chrom2, start2, end2, strand2 = row['chrom'], int(row['start']), int(row['end']), row['strand']
                if strand1 == '+':
                    ori1 = '+'
                    pos1 = end1
                else:
                    ori1 = '-'
                    pos1 = start1
                if strand2 == '+':
                    ori2 = '-'
                    pos2 = start2
                else:
                    ori2 = '+'
                    pos2 = end2
                tra = ((chrom1, pos1, ori1), (chrom2, pos2, ori2))
                self.list.append(tra)
                

class Segments:
    def __init__(self, df):
        self.df = df.copy()
        self.list = []
        self.get_list()

    def get_list(self):
        for qname, qdf in self.df.groupby('qname'):
            n_fragments = qdf.shape[0]
            if n_fragments <= 2: continue # don't count segments when none
            for rix, row in qdf.iterrows():
                if rix == 0 or rix == n_fragments-1: continue
                chrom, start, end = row['chrom'], int(row['start']), int(row['end'])
                segment = (chrom, start, end)
                self.list.append(segment)


def get_breakpoint_seqs(chrom, pos, margin, genome):
    linear_chroms = set(['chr'+str(c) for c in range(1, 22+1)] + ['chrX', 'chrY'])
    vector_lengths = { # GLOBAL - I don't want to waste time automating this...
        'GFPVector': 8725,
        'PiggyBacVector': 9794,
        'PGBD5Vector': 10218,
        'puro-GFP-PGBD5_seq': 10218,
        'DelPBEF1NeoTransposon': 6796,
        'PBEF1NeoTransposon': 6894,
    }
    assert chrom in genome, chrom
    # assert pos in genome[chrom], pos
    zpos = pos-1 # zero-based
    chrom_len = len(genome[chrom])
    start = zpos - margin
    end = zpos + margin
    if chrom in linear_chroms:
        start = max(0, start)
        end = min(chrom_len, end)
    
    if start < 0:
        upstream = genome[chrom][start:].seq.upper() + genome[chrom][:zpos].seq.upper() # upstream of 0
    else:
        upstream = genome[chrom][start : zpos].seq.upper() # upstream
    if end >= chrom_len:
        residual = end - chrom_len
        downstream = genome[chrom][zpos:].seq.upper() + genome[chrom][:residual].seq.upper() # downstream over end
    else:
        downstream = genome[chrom][zpos : end].seq.upper() # downstream
    return upstream, downstream


class Breakpoint:
    chroms = [str(c) for c in range(1, 22+1)] + ['X', 'Y', 'M']
    chroms += ['chr'+c for c in chroms]
    def __init__(self, chrom, pos, orientation):
        self.chrom = chrom
        self.pos = pos
        self.ori = orientation
        self.upstream = None
        self.downstream = None
        self.seq_rearranged = None
        self.seq_removed = None

    def get_breakpoint_seqs(self, margin, genome):
        self.upstream, self.downstream = get_breakpoint_seqs(self.chrom, self.pos, margin, genome)
        if self.ori == '+':
            self.seq_rearranged = self.upstream
            self.seq_removed = self.downstream
        elif self.ori == '-':
            self.seq_rearranged = self.downstream
            self.seq_removed = self.upstream
        else:
            raise ValueError(f'self.ori = {self.ori}')

    def __repr__(self):
        return f'{self.chrom}:{self.pos}:{self.ori}'

    def __lt__(self, other):
        self_chrom_ix = self.chroms.index(self.chrom)
        other_chrom_ix = self.chroms.index(other.chrom)
        if self_chrom_ix < other_chrom_ix:
            return True
        elif self_chrom_ix == other_chrom_ix and self.pos < other.pos:
            return True
        return False


class BreakpointPair:
    def __init__(self, brk1, brk2):
        self.brk1 = brk1
        self.brk2 = brk2

    def __repr__(self):
        return f'{self.brk1.chrom}:{self.brk1.pos}:{self.brk1.ori}-{self.brk2.chrom}:{self.brk2.pos}:{self.brk2.ori}'




import yaml
import swalign
import pyfaidx 
import pytest
import pandas as pd

from ontmont.datatypes import (Breakpoint, BreakpointPair)
from ontmont.utils import reverse_complement
from ontmont.irs import (calc_distance_score, get_best_onesided_ir, align_two_sequences,
    get_onesided_ir, get_best_ir_within_breakpoints, get_best_ir_within_segment,
    calc_pval_onesided_ir, calc_pval_bilateral_ir, calc_pval_segmental_ir, calc_pval_holliday,
    overlaps, is_holliday_junction, get_best_holliday_junctions, get_breakpoint_pair_seq_data)

@pytest.fixture
def config():
    CONFIG = """
    palindrome:
        match: 1
        mismatch: -2
        gap_penalty: -3
        gap_extension_penalty: -1
        cutoff1: 6
        cutoff2: 8
        bp_near_breakpoint: 5
        random_seed: 42
        n_iter: 1000
        margin: 30

    holliday:
        match: 1
        mismatch: -2
        gap_penalty: -2
        gap_extension_penalty: -1
        distance_cutoff1: 1
        distance_cutoff2: 3
        score_cutoff: 5
        random_seed: 42
        n_iter: 1000
        margin: 15

    reference:
        '1': /data1/shahs3/users/chois7/projects/transposon/resources/GRCh38_PBEF1_GFPVector.fa
        '2': /data1/shahs3/users/chois7/projects/transposon/resources/GRCh38_PBEF1_PGBD5Vector.fa
        '3': /data1/shahs3/users/chois7/projects/transposon/resources/GRCh38_PBEF1_PiggyBacVector.fa
        '5': /data1/shahs3/users/chois7/projects/transposon/resources/GRCh38_DelPBEF1_PGBD5Vector.fa
    """
    return yaml.safe_load(CONFIG)


@pytest.fixture
def sw_palindrome(config):
    scoring = swalign.NucleotideScoringMatrix(float(config['palindrome']['match']), float(config['palindrome']['mismatch']))
    sw = swalign.LocalAlignment(scoring, globalalign=False,
        gap_penalty=float(config['palindrome']['gap_penalty']), 
        gap_extension_penalty=float(config['palindrome']['gap_extension_penalty']))
    return sw

@pytest.fixture
def sw_holliday(config):
    scoring = swalign.NucleotideScoringMatrix(float(config['holliday']['match']), float(config['holliday']['mismatch']))
    sw = swalign.LocalAlignment(scoring, globalalign=False,
        gap_penalty=float(config['holliday']['gap_penalty']), 
        gap_extension_penalty=float(config['holliday']['gap_extension_penalty']))
    return sw

# TESTS

def test_reverse_complement():
    seq_revs = {'ACGGT':'ACCGT', 'GNTTCA': 'TGAA.C'}
    for seq, rev in seq_revs.items():
        assert reverse_complement(seq) == rev, (seq, rev)

def test_calc_distance_score():
    assert calc_distance_score(dist1=0, dist2=0, dist_cutoff1=1, dist_cutoff2=3) == 6
    assert calc_distance_score(dist1=1, dist2=3, dist_cutoff1=1, dist_cutoff2=3) == 4
    assert calc_distance_score(dist1=2, dist2=3, dist_cutoff1=1, dist_cutoff2=3) == 2
    assert calc_distance_score(dist1=2, dist2=4, dist_cutoff1=1, dist_cutoff2=3) == 1
    assert calc_distance_score(dist1=4, dist2=4, dist_cutoff1=1, dist_cutoff2=3) == 0


def test_align_two_sequences(config, sw_palindrome):
    genome = pyfaidx.Fasta(config['reference']['3'])

    brk = Breakpoint('PBEF1NeoTransposon', 1478, '+')
    brk.get_breakpoint_seqs(margin=50, genome=genome)
    seq1 = brk.upstream
    seq2 = brk.downstream
    sw = sw_palindrome
    aln = align_two_sequences(seq1, seq2, sw, rc=True)
    up_start, up_end = aln.up_coords
    down_start, down_end = aln.down_coords
    assert aln.seq1[up_start:up_end] == 'ATGCA', aln.seq1[up_start:up_end]
    assert aln.seq2[down_start:down_end] == reverse_complement('ATGCA'), aln.dump()

    seq1 = 'TTAAAAAAAT'
    seq2 = 'CAAAAAAACC'
    aln = align_two_sequences(seq1, seq2, sw, rc=False) # not rc
    assert aln.up_coords == (2, 2+7), aln.up_coords
    assert aln.down_coords == (1, 1+7), aln.down_coords
    up_start, up_end = aln.up_coords
    down_start, down_end = aln.down_coords
    assert aln.seq1[up_start:up_end] == 'AAAAAAA', aln.seq1[up_start:up_end]
    assert aln.seq2[down_start:down_end] == 'AAAAAAA', aln.seq2[down_start:down_end]


def test_get_best_onesided_ir(sw_palindrome):
    seq = 'TTTTTTTTTTTTAGCACTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTGTGCTAA'
    sw = sw_palindrome
    direction = 'up' # sequence is upstream of a breakpoint

    aln = get_best_onesided_ir(seq, direction, sw, margins=[15, 30, 60])
    assert aln.score == 7, aln.dump()
    assert aln.seq[aln.up_coords[0]:aln.up_coords[1]] == 'TTAGCAC', aln.dump()
    assert aln.seq[aln.down_coords[0]:aln.down_coords[1]] == 'GTGCTAA', aln.dump()

    aln = get_best_onesided_ir(seq, direction, sw, margins=[15, 30])
    assert aln.score == 2, aln.dump()
    assert aln.seq[aln.up_coords[0]:aln.up_coords[1]] == 'TT', aln.dump()
    assert aln.seq[aln.down_coords[0]:aln.down_coords[1]] == 'AA', aln.dump()

    seq = 'GAGCTTGTTGTACAG'
    direction = 'up'
    aln = get_best_onesided_ir(seq, direction, sw, margins=[15]) # , ((up_start, up_end), (down_start, down_end))
    assert aln == None, aln

def test_get_onesided_ir(sw_palindrome):
    sw = sw_palindrome
    seq = 'TTAGCACTTTTTTTTTTTTTGTGCTAAAAA'
    aln = get_onesided_ir(seq, sw) # , ((up_start, up_end), (down_start, down_end))
    up_start, up_end = aln.up_coords
    down_start, down_end = aln.down_coords
    assert up_start == 0, aln.dump()
    assert up_end == 7, aln.dump()
    assert down_start == 20, aln.dump()
    assert down_end == 27, aln.dump()
    assert seq[up_start:up_end] == reverse_complement(seq[down_start:down_end]) == 'TTAGCAC', (
        seq[up_start:up_end], seq[down_start:down_end]
    )
    assert len(seq) - down_end == 3, aln.dump() 

def test_calc_pval_onesided_ir_1(sw_palindrome):
    sw = sw_palindrome
    seq = 'GGGGAAAAATTTTTGGGGGGGAAAAATTTTT'
    direction = 'up'
    aln = get_onesided_ir(seq, sw)
    src_score = aln.score
    pval = calc_pval_onesided_ir(seq, sw, direction, src_score, dist_cutoff=5, n_iter=100, random_seed=42)
    assert src_score == 10.0, src_score
    assert pval == 0.0, pval

def test_calc_pval_onesided_ir_2(sw_palindrome):
    sw = sw_palindrome
    seq = 'CCCTTGGGGAAGGGTGGGGTGACGCGGGGG' # downstream
    direction = 'down'
    aln = get_onesided_ir(seq, sw)
    src_score = aln.score
    pval = calc_pval_onesided_ir(seq, sw, direction, src_score, dist_cutoff=5, n_iter=100, random_seed=42)
    assert src_score == 5.0, src_score
    assert pval == 0.01, pval

def test_calc_pval_onesided_ir_3(sw_palindrome):
    sw = sw_palindrome
    seq = 'GGGGGCGCAGTGGGGTGGGAAGGGGTTCCC' # upstream, reverse of seq from _2
    direction = 'up'
    aln = get_onesided_ir(seq, sw)
    src_score = aln.score
    pval = calc_pval_onesided_ir(seq, sw, direction, src_score, dist_cutoff=5, n_iter=100, random_seed=42)
    assert src_score == 5.0, src_score
    assert pval == 0.01, pval

def test_overlaps():
    seg1 = (8, 14)
    seg2 = (8, 14)
    assert overlaps(seg1, seg2), (seg1, seg2)
    seg1 = (8, 14)
    seg2 = (13, 15)
    assert overlaps(seg1, seg2), (seg1, seg2)

def test_get_best_ir_within_breakpoints(sw_palindrome):
    seq1 = 'TTTTTTTTTTTTTCGCGAGCCGTCGTTGCGCTTGTGCTAA'
    seq2 = 'TTAGCACTTTTTTTTTTTTTTTTTTTTTTTAAAAAAAAAA'
    sw = sw_palindrome

    aln = get_best_ir_within_breakpoints(seq1, seq2, sw, dist_cutoff1=1, dist_cutoff2=3, margins=[15, 40])
    assert aln.score == 7, aln.dump()
    assert aln.seq1[aln.up_coords[0]:aln.up_coords[1]] == 'GTGCTAA', aln.dump()
    assert aln.seq2[aln.down_coords[0]:aln.down_coords[1]] == 'TTAGCAC', aln.dump()

def test_calc_pval_bilateral_ir_1(sw_palindrome):
    sw = sw_palindrome
    seq1 = 'AAAAAAAAAA'
    seq2 = 'TTTTTTTTTT'
    aln = align_two_sequences(seq1, seq2, sw, rc=True)
    src_score = aln.score
    pval = calc_pval_bilateral_ir(seq1, seq2, sw, src_score, dist_cutoff1=1, dist_cutoff2=3, n_iter=100, random_seed=42)
    assert pval == 1, pval

def test_calc_pval_bilateral_ir_2(sw_palindrome):
    sw = sw_palindrome
    seq1 = 'AAAAAAAAAA'
    seq2 = 'TTAAATTTTT'
    aln = align_two_sequences(seq1, seq2, sw, rc=True)
    src_score = aln.score
    pval = calc_pval_bilateral_ir(seq1, seq2, sw, src_score, dist_cutoff1=100, dist_cutoff2=10, n_iter=10, random_seed=42)
    assert pval == 0.3, pval

def test_get_best_ir_within_segment(config, sw_palindrome):
    genome = pyfaidx.Fasta(config['reference']['3'])
    sw = sw_palindrome
    margin = 50
    chrom1, pos1, ori1 = 'PBEF1NeoTransposon', 1478, '-'
    chrom2, pos2, ori2 = 'PBEF1NeoTransposon', 4996, '+'
    brk1 = Breakpoint(chrom1, pos1, ori1)
    brk2 = Breakpoint(chrom2, pos2, ori2)
    pair = BreakpointPair(brk1, brk2)
    pair.brk1.get_breakpoint_seqs(margin, genome)
    pair.brk2.get_breakpoint_seqs(margin, genome)
    
    pair = get_best_ir_within_segment(pair, sw, genome, margins=[margin])
    assert pair.aln_segment.flag_segment, pair.aln_segment.dump()
    assert pair.aln_segment.score == 17, pair.aln_segment.dump()
    start1, end1 = pair.aln_segment.up_coords
    assert pair.brk1.seq_rearranged[start1:end1] == 'TTAACCCTAGAAAGATAGTC', pair.brk1.seq_rearranged[start1:end1]
    start2, end2 = pair.aln_segment.down_coords
    assert pair.brk2.seq_rearranged[start2:end2] == 'GATTATCTTTCTAGGGTTAA', pair.brk2.seq_rearranged[start2:end2]

def test_calc_pval_segmental_ir_1(sw_palindrome):
    sw = sw_palindrome
    seq1 = 'AAAAAAAAAA'
    seq2 = 'TTTTTTTTTT'
    ori1 = '-'
    ori2 = '+'
    aln = align_two_sequences(seq1, seq2, sw, rc=True)
    aln.dump()
    src_score = aln.score
    pval = calc_pval_segmental_ir(seq1, seq2, ori1, ori2, sw, src_score, 
                                  dist_cutoff1=1, dist_cutoff2=3, n_iter=100, random_seed=42)
    assert pval == 1, pval # shuffling sequence doesn't affect aln

def test_calc_pval_segmental_ir_2(sw_palindrome):
    sw = sw_palindrome
    seq1 = 'AAAAGGCGG'
    seq2 = 'TTAAATTTT'
    ori1 = '-'
    ori2 = '+'
    aln = align_two_sequences(seq1, seq2, sw, rc=True)
    src_score = aln.score
    pval = calc_pval_segmental_ir(seq1, seq2, ori1, ori2, sw, src_score, 
                                  dist_cutoff1=1000, dist_cutoff2=10, n_iter=10, random_seed=42)
    assert pval == 0.1, pval

def test_is_holliday_junction(sw_palindrome):
    score_cutoff = 4
    dist_sum_cutoff = 4 # HARD CODED
    sw = sw_palindrome
    seq1 = 'AACCGGTTACGTACGT'
    seq2 = 'TTTTACGTAACCGGTA'
    negative1, negative2 = False, False
    aln = align_two_sequences(seq1, seq2, sw, rc=False)
    aln = is_holliday_junction(aln, negative1, negative2, 
        dist_cutoff1=1, dist_cutoff2=9, dist_sum_cutoff=dist_sum_cutoff, score_cutoff=score_cutoff)
    assert aln.flag_holliday, aln.dump()

def test_calc_pval_holliday_u1u2(sw_holliday):
    sw = sw_holliday
    seq1 = 'TTACGC'+'CAAGCGCGC'
    seq2 = 'GCACAG'+'CAAGCGCGC'
    negative1 = False # downstream ^ rc?
    negative2 = False # downstream ^ rc?
    aln = align_two_sequences(seq1, seq2, sw, rc=False) # no rc
    assert aln.score == 9, aln.dump()
    src_score = aln.score
    pval = calc_pval_holliday(seq1, seq2, negative1, negative2, src_score, sw,
                              dist_cutoff1=1, dist_cutoff2=3, n_iter=100, random_seed=42)
    assert pval == 0.0, pval

def test_calc_pval_holliday_d1d2(sw_holliday):
    sw = sw_holliday
    seq1 = 'CAAGC'+'TTACGC'
    seq2 = 'CAAGC'+'GCACAG'
    negative1 = True # downstream ^ rc?
    negative2 = True # downstream ^ rc?
    aln = align_two_sequences(seq1, seq2, sw, rc=False) # no rc
    assert aln.score == 5, aln.dump()
    src_score = aln.score
    pval = calc_pval_holliday(seq1, seq2, negative1, negative2, src_score, sw,
                              dist_cutoff1=1, dist_cutoff2=3, n_iter=100, random_seed=42)
    assert pval == 0.02, pval

def test_calc_pval_holliday_u1d2r(sw_holliday):
    sw = sw_holliday
    seq1 = 'TTACGC'+'CAAG'
    seq2 = 'CAAG'+'GCACAG'
    negative1 = False # downstream ^ rc?
    negative2 = True # downstream ^ rc?
    aln = align_two_sequences(seq1, seq2, sw, rc=False) # no rc
    aln = is_holliday_junction(aln, negative1, negative2, 
                               dist_cutoff1=1, dist_cutoff2=3, 
                               dist_sum_cutoff=4, score_cutoff=4)
    assert aln.flag_holliday, aln.dump()
    assert aln.score == 4, aln.dump()
    src_score = aln.score
    pval = calc_pval_holliday(seq1, seq2, negative1, negative2, src_score, sw,
                              dist_cutoff1=1, dist_cutoff2=3, n_iter=100, random_seed=42)
    assert pval == 0.01, pval

def test_calc_pval_holliday_d1u2r(sw_holliday):
    sw = sw_holliday
    seq1 = 'CAAG'+'GCACAG'
    seq2 = 'TTACGC'+'CAAG'
    negative1 = True # downstream ^ rc?
    negative2 = False # downstream ^ rc?
    aln = align_two_sequences(seq1, seq2, sw, rc=False) # no rc
    aln = is_holliday_junction(aln, negative1, negative2, 
                               dist_cutoff1=1, dist_cutoff2=3, 
                               dist_sum_cutoff=4, score_cutoff=4)
    assert aln.flag_holliday, aln.dump()
    assert aln.score == 4, aln.dump()
    src_score = aln.score
    pval = calc_pval_holliday(seq1, seq2, negative1, negative2, src_score, sw,
                              dist_cutoff1=1, dist_cutoff2=3, n_iter=100, random_seed=42)
    assert pval == 0.01, pval

def test_get_breakpoint_pair_seq_data():
    brk1 = Breakpoint('chr1', 1000, '+')
    brk2 = Breakpoint('chrX', 2000, '-')
    brk1.upstream = 'AAAAA' # up1
    brk1.downstream = 'CCCCG' # dn1
    brk2.upstream = 'GCGGG' # up2
    brk2.downstream = 'ATTTT' # dn2
    pair = BreakpointPair(brk1, brk2)
    seqs = get_breakpoint_pair_seq_data(pair)
    assert seqs['u1'] == 'AAAAA', seqs['u1']
    assert seqs['u2'] == 'GCGGG', seqs['u2']
    assert seqs['d1'] == 'CCCCG', seqs['d1']
    assert seqs['d2'] == 'ATTTT', seqs['d2']
    assert seqs['u2r'] == 'GCGGG', seqs['u2r']
    assert seqs['d2r'] == 'ATTTT', seqs['d2r']


def test_get_best_holliday_junction_1(config, sw_holliday): 
    genome = pyfaidx.Fasta(config['reference']['2'])
    sw = sw_holliday

    chrom1, pos1, ori1 = 'PGBD5Vector', 1012, '+'
    chrom2, pos2, ori2 = 'PGBD5Vector', 1980, '+'
    brk1 = Breakpoint(chrom1, pos1, ori1)
    brk2 = Breakpoint(chrom2, pos2, ori2)
    pair = BreakpointPair(brk1, brk2)
    pair = get_best_holliday_junctions(pair, sw, genome)
    alns = pair.alns
    assert alns['u1_u2'].score == 6, alns['u1_u2'].dump()
    start1, end1 = alns['u1_u2'].up_coords
    assert (start1, end1) == (6, 15), (start1, end1)
    assert alns['u1_u2'].seq1[start1:end1] == 'CAAGCGCGC', alns['u1_u2'].seq1
    start2, end2 = alns['u1_u2'].down_coords
    assert (start2, end2) == (6, 14), (start2, end2)
    assert alns['u1_u2'].seq2[start2:end2] == 'CAAGCG-GC'.replace('-', ''), alns['u1_u2'].seq1

def test_get_best_holliday_junction_2(config, sw_holliday): 
    genome = pyfaidx.Fasta(config['reference']['2'])
    sw = sw_holliday

    chrom1, pos1, ori1 = 'PGBD5Vector', 1285, '-'
    chrom2, pos2, ori2 = 'PGBD5Vector', 3421, '-'
    brk1 = Breakpoint(chrom1, pos1, ori1)
    brk2 = Breakpoint(chrom2, pos2, ori2)
    pair = BreakpointPair(brk1, brk2)
    pair = get_best_holliday_junctions(pair, sw, genome)
    alns = pair.alns
    
    assert alns['u1_u2'] == None, alns['u1_u2']
    assert alns['d1_d2'] == None, alns['d1_d2']
    assert alns['u1_d2r'] == None, alns['u1_d2r']
    assert alns['d1_u2r'].score == 5, alns['d1_u2r'].dump()

    start1, end1 = alns['d1_u2r'].up_coords
    assert (start1, end1) == (0, 5), (start1, end1)
    assert pair.brk1.downstream[start1:end1] == 'GGGTC', alns['d1_u2r'].dump() # pair.brk1.downstream #pair.brk1.upstream[start1:end1]
    start2, end2 = alns['d1_u2r'].down_coords
    assert (start2, end2) == (9, 14), (start2, end2)
    assert reverse_complement(pair.brk2.upstream[-15:][start2:end2]) == 'GGGTC', pair.brk2.upstream #pair.brk1.upstream[start1:end1]



import pytest

import yaml
import pyfaidx
import pandas as pd
import numpy as np

from ontmont.utils import remove_duplicates_from_tra_table, filter_sv_with_breakpoint_at_contig_ends


def test_remove_duplicates_from_tra_tables():
    tra_df_columns = ['chrom1', 'pos1', 'ori1', 'chrom2', 'pos2', 'ori2', 'support', 
                      'u1_u2_score', 'd1_d2_score', 'u1_d2r_score', 'd1_u2r_score',
                      'u1_u2_pvalue', 'd1_d2_pvalue', 'u1_d2r_pvalue', 'd1_u2r_pvalue']
    data = [
        ['chr1', 100, '+', 'chr2', 200, '-', 1,  0, 0, 3, 4, np.nan, np.nan, 0.01, 0.02],
        ['chr2', 200, '-', 'chr1', 100, '+', 1,  0, 1, 4, 3, np.nan, 0.3, 0.02, 0.01],
    ]
    inputdf = pd.DataFrame(data=data, index=range(0, 2), columns=tra_df_columns)
    outputdf = remove_duplicates_from_tra_table(inputdf)
    
    out_data = [
        ['chr1', 100, '+', 'chr2', 200, '-', 2,  0, 1, 3, 4, np.nan, 0.3, 0.01, 0.02],
    ]
    val = pd.DataFrame(data=out_data, index=range(0, 1), columns=tra_df_columns)
    assert outputdf.shape[0] == val.shape[0], outputdf
    assert np.isnan(outputdf.squeeze()['u1_u2_pvalue']), outputdf.squeeze()['u1_u2_pvalue']
    assert (outputdf.dropna(axis=1) != val.dropna(axis=1)).sum().sum() == 0, outputdf


def test_filter_sv_with_breakpoint_at_contig_ends():
    df = pd.DataFrame({
    0: {
        'chrom': 'PBEF1NeoTransposon',
        'pos1': 1,
        'pos2': 6895,
        'support': 4,
        'segment_score': 9.0,
        'segment_pvalue': 0.0,
    }}).T
    outputdf = filter_sv_with_breakpoint_at_contig_ends(df)
    assert outputdf.shape[0] == 0, outputdf


import random

import pandas as pd
import numpy as np

from .datatypes import Breakpoint, BreakpointChain, SplitAlignment

def remove_duplicates_from_tra_table(tra_df):
    df = pd.DataFrame(columns = tra_df.columns)
    
    vectors = [
        'DelPBEF1NeoTransposon',
        'GFPVector',
        'PBEF1NeoTransposon',
        'PGBD5Vector',
        'PiggyBacVector',
        'puro-GFP-PGBD5_seq'
    ]
    chrom_order = ['chr'+str(c) for c in range(1, 22)] + ['chrX', 'chrY'] + vectors
    
    for rix, row in tra_df.iterrows():
        chrom1, pos1, ori1 = row['chrom1'], row['pos1'], row['ori1']
        chrom2, pos2, ori2 = row['chrom2'], row['pos2'], row['ori2']
        if chrom1 not in chrom_order or chrom2 not in chrom_order:
            continue
        support = row['support']
        u1_u2_score, d1_d2_score = row['u1_u2_score'], row['d1_d2_score']
        u1_d2r_score, d1_u2r_score = row['u1_d2r_score'], row['d1_u2r_score']
        u1_u2_pvalue, d1_d2_pvalue = row['u1_u2_pvalue'], row['d1_d2_pvalue']
        u1_d2r_pvalue, d1_u2r_pvalue = row['u1_d2r_pvalue'], row['d1_u2r_pvalue']
        flag_reverse = False
        if chrom1 != chrom2:
            if chrom_order.index(chrom1) > chrom_order.index(chrom2):
                flag_reverse = True
        else:
            if pos1 > pos2:
                flag_reverse = True
        if flag_reverse:
            chrom1, pos1, ori1, chrom2, pos2, ori2 = chrom2, pos2, ori2, chrom1, pos1, ori1
            u1_d2r_score, d1_u2r_score = d1_u2r_score, u1_d2r_score
            u1_d2r_pvalue, d1_u2r_pvalue = d1_u2r_pvalue, u1_d2r_pvalue
        field = [chrom1, pos1, ori1, chrom2, pos2, ori2, support, 
            u1_u2_score, d1_d2_score, u1_d2r_score, d1_u2r_score,
            u1_u2_pvalue, d1_d2_pvalue, u1_d2r_pvalue, d1_u2r_pvalue]
        df.loc[df.shape[0]] = field
    
    ix_cols = ['chrom1', 'pos1', 'ori1', 'chrom2', 'pos2', 'ori2']
    df = df.groupby(ix_cols).agg({
        'support':['sum'], 
        'u1_u2_score':['max'], 'd1_d2_score':['max'], 'u1_d2r_score':['max'], 'd1_u2r_score':['max'],
        'u1_u2_pvalue':['min'], 'd1_d2_pvalue':['min'], 'u1_d2r_pvalue':['min'], 'd1_u2r_pvalue':['min']
    })
    df.columns = df.columns.droplevel(1)
    df = df.reset_index()
    return df


def enumerate_breakpoints(df):
    df = df.reset_index(drop=True)
    ix_start, ix_end = 0, df.shape[0] - 1
    brks = BreakpointChain([])
    for rix, row in df.iterrows():
        chrom = row['chrom']
        start = int(row['start'])
        end = int(row['end'])
        strand = row['strand']
        # match = int(row['match'])
        if rix == ix_start:
            ori = '+' if strand == '+' else '-'
            pos = end if strand == '+' else start
            brk = Breakpoint(chrom, pos, ori)
            _brks = [brk]
        elif rix == ix_end:
            ori = '-' if strand == '+' else '+'
            pos = start if strand == '+' else end
            brk = Breakpoint(chrom, pos, ori)
            _brks = [brk]
        else:
            brk_start = Breakpoint(chrom, start, '-')
            brk_end = Breakpoint(chrom, end, '+')
            _brks = [brk_start, brk_end] if strand == '+' else [brk_end, brk_start]
        brks += _brks
    return brks


def filter_breakpoints_at_contig_ends(brk_df):
    vector_lengths = { # GLOBAL - I don't want to waste time automating this...
        'GFPVector': 8725,
        'PiggyBacVector': 9794,
        'PGBD5Vector': 10218,
        'puro-GFP-PGBD5_seq': 10218,
        'DelPBEF1NeoTransposon': 6796,
        'PBEF1NeoTransposon': 6894,
    }
    cols = brk_df.columns
    brk_df.loc[:, 'remove'] = False
    for rix, row in brk_df.iterrows():
        chrom = row['chrom']
        pos = row['pos']
        if chrom not in vector_lengths: continue # only remove circularities
        vector_length = vector_lengths[chrom]
        
        remove = pos <= 5 or abs(pos - vector_length) <= 5
        brk_df.loc[rix, 'remove'] = remove
    brk_df = brk_df[~brk_df['remove']][cols]
    return brk_df


def filter_sv_with_breakpoint_at_contig_ends(df):
    """
    Filters structural variant (SV) records to exclude those with breakpoints too close to the ends of contigs.

    This function calculates the distance of each SV breakpoint from the nearest end of its respective contig.
    SVs with either breakpoint within 3 bases of a contig end are excluded from the results. The function uses
    a predefined dictionary of contig lengths to determine the distances from the contig ends.

    Parameters:
    - df (pandas.DataFrame): A DataFrame containing SV records. Must include 'pos1' and 'chrom1' columns,
      where 'pos1' is the position of the breakpoint and 'chrom1' is the name of the contig.

    Returns:
    - pandas.DataFrame: A filtered DataFrame excluding SVs with breakpoints within 3 bases of the contig ends.
    """
    vector_lengths = { # GLOBAL - I don't want to waste time automating this...
        'GFPVector': 8725,
        'PiggyBacVector': 9794,
        'PGBD5Vector': 10218,
        'puro-GFP-PGBD5_seq': 10218,
        'DelPBEF1NeoTransposon': 6796,
        'PBEF1NeoTransposon': 6894,
    }
    df_cols = df.columns
    for chrom_col in ('chrom1', 'chrom2'):
        if chrom_col not in df_cols:
            assert 'chrom' in df_cols, df_cols
            df[chrom_col] = df['chrom']
            
    df['dist1s'] = np.abs(df['pos1'] - 1)
    df['dist1e'] = np.abs(df['pos1'] - 1 - df['chrom1'].map(vector_lengths))
    df['dist2s'] = np.abs(df['pos2'] - 1)
    df['dist2e'] = np.abs(df['pos2'] - 1 - df['chrom2'].map(vector_lengths))
    df[['dist1s', 'dist1e', 'dist2s', 'dist2e']] = df[['dist1s', 'dist1e', 'dist2s', 'dist2e']].astype(float).fillna(np.inf)
    df = df[(df['dist1s'] >= 3) & (df['dist1e'] >= 3) & (df['dist2s'] >= 3) & (df['dist2e'] >= 3)]
    df = df[df_cols]
    return df


def get_secondaries(read):
    """
    Extracts secondary alignment information from a read's tags.

    This function parses the tags of a given read to find the 'SA' tag, which contains information about
    secondary alignments. The 'SA' tag is expected to be a semicolon-separated string containing details
    of each secondary alignment. These details are then split into a list, with each element representing
    one secondary alignment.

    Parameters:
    - read (pysam.AlignedSegment): A single read object from a SAM/BAM file, as provided by pysam.

    Returns:
    - list: A list of strings, where each string contains the comma-separated details of a secondary
      alignment as specified in the SAM format specification. Returns an empty list if no secondary
      alignments are found.
    """
    secondaries = [a[1].split(';') for a in read.tags if a[0] == 'SA']
    if len(secondaries) > 0:
        secondaries = secondaries[0]
        secondaries = list(filter(lambda x: x!='', secondaries))
        return secondaries
    return []


def get_chromosomes_to_process(bam, drop_expression_vectors=False):
    """
    Identifies and returns a list of contigs from a BAM file that are specified in a predefined list and have at least one read aligned.

    Parameters:
    - bam (pysam.AlignmentFile): An object representing an opened BAM file for reading.

    Returns:
    - list: A list of contig names that are both present in the predefined list of vector contigs and have at least one read aligned in the BAM file.

    This function filters for specific vector contigs, checking each for the presence of aligned reads before adding it to the list of contigs to process.
    """
    vector_contigs = ['GFPVector', 'PiggyBacVector', 'PGBD5Vector', 
        'puro-GFP-PGBD5_seq', 'DelPBEF1NeoTransposon', 'PBEF1NeoTransposon']
    if drop_expression_vectors:
        vector_contigs = ['DelPBEF1NeoTransposon', 'PBEF1NeoTransposon']
    chroms_proc = []
    for contig in bam.references:
        if contig not in vector_contigs: continue
        if bam.count(contig=contig):
            print(contig)
            chroms_proc.append(contig)
    return chroms_proc


def extract_split_alignments(bam, chroms_proc):
    """
    Extracts and organizes split alignments from a BAM file for specified chromosomes.

    This function iterates over each chromosome in `chroms_proc`, fetching reads from the BAM file. 
    It focuses on reads that have secondary alignments (indicative of chimeric reads) and constructs 
    a list of `SplitAlignment` objects representing the primary and secondary alignments. Each 
    `SplitAlignment` contains the CIGAR string, query name, reference name, position (1-based), and strand 
    of the alignment.

    Parameters:
    - bam (pysam.AlignmentFile): An open BAM file for reading alignments.
    - chroms_proc (list of str): A list of chromosome names to process.

    Returns:
    - list of SplitAlignment: A list of `SplitAlignment` objects representing the extracted alignments.

    Note:
    - This function assumes the existence of a `SplitAlignment` class and a `get_secondaries` function 
      that extracts secondary alignments from a read.
    - Reads without secondary alignments are skipped.
    """
    alignments = []
    for chrom in chroms_proc:
        for read in bam.fetch(chrom):
            secondary_list = get_secondaries(read)
            if len(secondary_list) < 1: # only consider reads with secondaries (chimeric reads)
                continue # return None
            strand = '-' if read.is_reverse else '+'
            tpos = read.pos + 1
            cigarstring = read.cigarstring
            qname = read.qname
            refname = read.reference_name
            alignment = SplitAlignment(cigarstring, qname, refname, tpos, strand)
            alignments.append(alignment)
    
            for _, secondary in enumerate(secondary_list):
                refname_s, tpos_s, strand_s, cigarstring_s, _, _ = secondary.split(',')
                tpos_s = int(tpos_s)
                alignment = SplitAlignment(cigarstring_s, qname, refname_s, tpos_s, strand_s)
                alignments.append(alignment)
    return alignments

def make_split_read_table(alignments):
    """
    Creates a pandas DataFrame from a list of split read alignments.

    This function takes a list of alignment objects and transforms it into a structured pandas DataFrame. 
    Each row in the DataFrame represents a split read alignment with detailed information such as query name, 
    chromosome, start and end positions, strand, clipping information, and match length. The DataFrame is 
    sorted by query name and the position of the first clip, and duplicates are removed to ensure uniqueness 
    of entries.

    Parameters:
    - alignments (list): A list of alignment objects, each containing attributes like read_name, refname, 
      start, end, strand, clip1, clip2, match, and pclip1.

    Returns:
    - pandas.DataFrame: A DataFrame with columns ['qname', 'chrom', 'start', 'end', 'strand', 'clip1', 
      'match', 'clip2', 'pclip1'], sorted by 'qname' and 'pclip1', and without duplicates.
    """
    sa_cols = ['qname', 'chrom', 'start', 'end', 'strand', 'clip1', 'match', 'clip2', 'pclip1']
    df = pd.DataFrame(columns=sa_cols)
    for alignment in alignments:
        qname = alignment.read_name
        chrom = alignment.refname
        start = alignment.start
        end = alignment.end
        strand = alignment.strand
        clip1 = alignment.clip1
        clip2 = alignment.clip2
        match = alignment.match
        pclip1 = alignment.pclip1
        field = [qname, chrom, start, end, strand, clip1, match, clip2, pclip1]
        df.loc[df.shape[0]] = field
    df.drop_duplicates(inplace=True)
    df.sort_values(by=['qname', 'pclip1'], inplace=True)
    return df.reset_index(drop=True)

def is_breakpoints_not_sorted(chrom1, pos1, chrom2, pos2, chrom_order):
    """
    Check whether two breakpoints are ordered correctly or not
    """
    if chrom_order.index(chrom1) > chrom_order.index(chrom2):
        return True
    if chrom_order.index(chrom1) == chrom_order.index(chrom2) and pos1 > pos2:
        return True
    return False

def reverse_complement(seq):
    comp = {'A':'T', 'C':'G', 'G':'C', 'T':'A', 'N':'.'}
    revcmp = ''.join([comp[b] for b in seq[::-1]])
    return revcmp

def shuffle_seq(seq):
    seq_list = [s for s in seq]
    random.shuffle(seq_list)
    shuffled_seq = ''.join(seq_list)
    return shuffled_seq